unit CDInfo;

interface

uses
  MMSystem, Windows, Classes, Messages, Controls, ExtCtrls;

type
  // declares my CD TOC
  TOC = record
    min: LongWord;
    sec: LongWord;
    frames: LongWord;
  end;

  TCDInfo = class(TComponent)
  private
    mciStatusParms : MCI_STATUS_PARMS;
    FOnEject: TNotifyEvent;
    FOnInsert: TNotifyEvent;
    FTimer : TTimer;
    FReady : Boolean;
    FDeviceID : integer;
    procedure SendCommand(uMessage: UINT; dwCommand, dwParam : LongWord);
    function GetStatusItem (AStatusItem : LongWord) : LongWord;
    procedure RealSeek(APosn: TOC);
    procedure TimerOnTimer(Sender : TObject);
  public
    constructor Create(AOwner : Tcomponent); override;
    destructor Destroy; override;

    // information methods
    function TotalTrackCount : LongWord;
    function TrackPosition (ATrack : integer) : TOC;
    function TrackLength (ATrack : integer) : TOC;
    function TotalCDLength : LongWord;
    function CDDBDiscId : LongWord;
    function CDDBDiscIDHex : string;
    class function TOCToFrames (ATOC : TOC) : LongWord;
    class function MSFToTOC(AFrames: LongWord): TOC;
    class function FramesToTOC(AFrames: LongWord): TOC;
    class function TOCToString(ATOC : TOC) : string;

    // play functionality
    procedure Play;
    procedure Pause;
    procedure Stop;
    procedure Eject;
    procedure Next;
    procedure Previous;
    procedure GotoTrack(ATrack : integer);
    procedure Seek(APosn: TOC);
    function Ready : Boolean;
    function Playing : Boolean;
    function CurrentTrack : LongWord;
    function ElapsedDiscTime : TOC;
    function ElapsedTrackTime : TOC;
    procedure SeachForward;
    procedure SearchBackwards;
  published
    property OnEject : TNotifyEvent read FOnEject write FOnEject;
    property OnInsert : TNotifyEvent read FOnInsert write FOnInsert;
  end;

procedure Register;

////////////////////////////////////////////////////////////////////////////////

implementation

uses
  SysUtils, Math, Dialogs;

const
	CSearchDelta = 150;

procedure Register;
begin
  RegisterComponents('DoogalSoft', [TCDInfo]);
end;

////////////////////////////////////////////////////////////////////////////////
{ TCDInfo }

constructor TCDInfo.Create(AOwner : TComponent);
var
	LmciOpenParms : MCI_OPEN_PARMS;
	LmciSetParms : MCI_SET_PARMS;
  LAliasGUID : TGUID;
begin
  inherited;

  FTimer := TTimer.Create(nil);
  FTimer.Interval := 500;
  FTimer.OnTimer := TimerOnTimer;
  FTimer.Enabled := True;

  // open MCI device
  LmciOpenParms.dwCallback := 0;
  LmciOpenParms.lpstrDeviceType := 'cdaudio';
  LmciOpenParms.lpstrElementName := '';
  LmciOpenParms.wDeviceID := 0;
  CreateGUID(LAliasGUID);
  LmciOpenParms.lpstrAlias := PChar(GUIDToString(LAliasGUID));
  SendCommand(MCI_OPEN, MCI_OPEN_TYPE or MCI_OPEN_ALIAS or MCI_OPEN_SHAREABLE,
    LongWord(@LmciOpenParms));
  FDeviceID := LmciOpenParms.wDeviceID;

  // set time format to minute/second/frame (MSF) format
  LmciSetParms.dwTimeFormat := MCI_FORMAT_MSF;
  SendCommand(MCI_SET, MCI_SET_TIME_FORMAT, LongWord(@LmciSetParms));

  FReady := Ready;
end;

////////////////////////////////////////////////////////////////////////////////

destructor TCDInfo.Destroy;
begin
  FTimer.Free;
  // close MCI device
  SendCommand(MCI_CLOSE, 0, 0);
  inherited;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.SendCommand(uMessage: UINT; dwCommand, dwParam : LongWord);
const
  CStrLength = 250;
var
  LError : MCIERROR;
  LErrorStr : string;
begin
  LError := mciSendCommand(FDeviceID, uMessage, dwCommand, dwParam);
  if LError <> 0 then
  begin
    SetLength(LErrorStr, CStrLength);
    mciGetErrorString(LError, PChar(LErrorStr), CStrLength);
    raise Exception.Create(LErrorStr);
  end;
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.TotalTrackCount : LongWord;
begin
	Result := GetStatusItem(MCI_STATUS_NUMBER_OF_TRACKS);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.TrackPosition (ATrack : integer) : TOC;
begin
  mciStatusParms.dwItem := MCI_STATUS_POSITION;
  mciStatusParms.dwTrack := ATrack;
  SendCommand(MCI_STATUS, MCI_STATUS_ITEM or MCI_TRACK, LongWord(@mciStatusParms));

  // save track position in MSF format
  Result := MSFToTOC(mciStatusParms.dwReturn);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.TrackLength (ATrack : integer) : TOC;
begin
  mciStatusParms.dwItem := MCI_STATUS_LENGTH;
  mciStatusParms.dwTrack := ATrack;
  SendCommand(MCI_STATUS, MCI_STATUS_ITEM or MCI_TRACK, LongWord(@mciStatusParms));

  // save track position in MSF format
  Result := MSFToTOC(mciStatusParms.dwReturn);
end;

////////////////////////////////////////////////////////////////////////////////

class function TCDInfo.TOCToFrames(ATOC: TOC): LongWord;
begin
  Result := (ATOC.min*60*75) + (ATOC.sec*75) + ATOC.frames;
end;

////////////////////////////////////////////////////////////////////////////////

class function TCDInfo.MSFToTOC(AFrames: LongWord): TOC;
begin
	Result.frames := MCI_MSF_FRAME(AFrames);
  Result.sec := MCI_MSF_SECOND(AFrames);
  Result.min := MCI_MSF_MINUTE(AFrames);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.TotalCDLength: LongWord;
var
  LLastTrack, LeadOutTrack : TOC;
  dwPos : LongWord;
begin
  /////////////////////////////////////
  // Compute lead-out track position //

  // get last track length
  LLastTrack := TrackLength(TotalTrackCount);
  // +1 to fix Windows MCI bug, according to cddb howto
  LLastTrack.frames := LLastTrack.frames + 1;

  // compute lead-out track position (in frame format)
  dwPos := TOCToFrames(TrackPosition(TotalTrackCount)) + TOCToFrames(LLastTrack);

  // save lead-out track position (in MSF format)
  LeadOutTrack := FramesToTOC(dwPos);

  // compute total CD length in seconds
  Result := ((LeadOutTrack.min * 60) + LeadOutTrack.sec) -
    ((TrackPosition(1).min * 60) + TrackPosition(1).sec);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.CDDBDiscIDHex: string;
begin
  Result := Format('%8.8x', [CDDBDiscId]);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.CDDBDiscId : LongWord;
var
  // general variables
  i : LongWord;

  // CD TOC information variables
  cdtoc : array[0..99] of TOC;

  // cddb discid computation variables
  j, n, cddb_sum : LongWord;

begin
  /////////////////
  // Read CD TOC //

  // for each track, get position (starting location)
  for i:= 1 to TotalTrackCount do
  begin
    cdtoc[i-1] := TrackPosition(i);
  end;

  /////////////////////
  // Compute DISC-ID //

  // For backward compatibility this algorithm must not change

  // cddb_discid
  i := 0; n := 0;
  while (i < TotalTrackCount) do
  begin
    // cddb_sum
    cddb_sum := 0;
    j := (cdtoc[i].min * 60) + cdtoc[i].sec;
    while (j > 0) do
    begin
      cddb_sum := cddb_sum + (j mod 10);
      j := j div 10;
    end;

    n := n + cddb_sum;
    Inc(i);
  end;

  // compute disc-id
  Result := ((n mod $ff) shl 24 or TotalCDLength shl 8 or TotalTrackCount);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Play;
begin
	SendCommand(MCI_PLAY, 0, 0);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Pause;
begin
	SendCommand(MCI_PAUSE, 0, 0);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Stop;
begin
	SendCommand(MCI_STOP, 0, 0);
  RealSeek(TrackPosition(1));
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.CurrentTrack: LongWord;
begin
  Result := GetStatusItem(MCI_STATUS_CURRENT_TRACK);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.ElapsedDiscTime : TOC;
begin
  Result := MSFToTOC(GetStatusItem(MCI_STATUS_POSITION));
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.GetStatusItem (AStatusItem : LongWord) : LongWord;
begin
  mciStatusParms.dwItem := AStatusItem;
  SendCommand(MCI_STATUS, MCI_STATUS_ITEM, LongWord(@mciStatusParms));

  Result := mciStatusParms.dwReturn;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Eject;
begin
  SendCommand(MCI_SET, MCI_SET_DOOR_OPEN, 0);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.Ready: Boolean;
begin
  Result := (GetStatusItem(MCI_STATUS_READY) = 1);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Next;
var
  LTrackToGoTo : integer;
begin
  // loop back to beginning if we're at the end
  if CurrentTrack = TotalTrackCount then
    LTrackToGoTo := 1
  else
    LTrackToGoTo := CurrentTrack+1;

  GotoTrack(LTrackToGoTo);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Previous;
var
	LTrackToGoTo : integer;
begin
  if CurrentTrack = 1 then
    LTrackToGoTo := TotalTrackCount
  else
    LTrackToGoTo := CurrentTrack-1;

  GotoTrack(LTrackToGoTo);
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.ElapsedTrackTime: TOC;
var
	LCurrentPos, LTrackPosition : integer;
begin
  LTrackPosition := TOCToFrames(TrackPosition(CurrentTrack));
  LCurrentPos := TOCToFrames(ElapsedDiscTime);

  Result := FramesToTOC(Max(0, LCurrentPos - LTrackPosition));
end;

////////////////////////////////////////////////////////////////////////////////

class function TCDInfo.FramesToTOC(AFrames : LongWord) : TOC;
begin
	Result.frames := AFrames mod 75;
  AFrames := AFrames div 75;
  Result.sec := AFrames mod 60;
  AFrames := AFrames div 60;
  Result.min := AFrames;
end;

////////////////////////////////////////////////////////////////////////////////

function TCDInfo.Playing: Boolean;
begin
  Result := (GetStatusItem(MCI_STATUS_MODE) = MCI_MODE_PLAY);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.GotoTrack(ATrack: integer);
var
  LTrackPos : TOC;
begin
  LTrackPos := TrackPosition(ATrack);
  Seek(LTrackPos);
end;

////////////////////////////////////////////////////////////////////////////////

class function TCDInfo.TOCToString(ATOC: TOC): string;
begin
  Result := IntToStr(ATOC.min) + ':' + Format('%.2d', [ATOC.sec]);
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.SeachForward;
var
	LCurrentPos : integer;
begin
  LCurrentPos := TOCTOFrames(ElapsedDiscTime);
  Inc(LCurrentPos, CSearchDelta);
  Seek(FramesToTOC(LCurrentPos));
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.SearchBackwards;
var
	LCurrentPos : integer;
begin
  LCurrentPos := TOCTOFrames(ElapsedDiscTime);
  Dec(LCurrentPos, CSearchDelta);
  Seek(FramesToTOC(LCurrentPos));
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.RealSeek(APosn: TOC);
var
	LSeekParm : TMCI_Seek_Parms;
begin
	LSeekParm.dwTo := mci_Make_MSF(APosn.min, APosn.sec, APosn.frames);
  SendCommand(MCI_SEEK, mci_To, Longint(@LSeekParm));
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.Seek(APosn: TOC);
var
	LPlaying : Boolean;
begin
  LPlaying := Playing;

  RealSeek(APosn);

  // restart if we were playing before
  if LPlaying then
  	Play;
end;

////////////////////////////////////////////////////////////////////////////////

procedure TCDInfo.TimerOnTimer(Sender: TObject);
var
  LReady : Boolean;
begin
  FTimer.Enabled := False;
  try
    LReady := Ready;

    if FReady and not LReady then
      if Assigned(FOnEject) then
        FOnEject(Self);

    if not FReady and LReady then
      if Assigned(FOnInsert) then
        FOnInsert(Self);

    FReady := LReady;
  finally
    FTimer.Enabled := True;
  end;
end;

////////////////////////////////////////////////////////////////////////////////

end.
